@@ -37,7 +37,7 @@ __author__ = "François Orieux"
 __copyright__ = "2021, François Orieux <francois.orieux@universite-paris-saclay.fr>"
 __credits__ = ["François Orieux"]
 __license__ = "GPL-3.0-or-later"
-__version__ = "0.12.1"
+__version__ = "0.13.1"
 __maintainer__ = "François Orieux"
 __email__ = "francois.orieux@universite-paris-saclay.fr"
 __status__ = "beta"
@@ -51,6 +51,7 @@ __all__ = [
     "mmmg",
     "mmcg",
     "lcg",
+    "pcg",
     "BaseObjective",
     "MixedObjective",
     "Objective",
@@ -161,6 +162,7 @@ def mmmg(  # pylint: disable=too-many-locals
     x0: array,  # pylint: disable=invalid-name
     tol: float = 1e-4,
     max_iter: int = 500,
+    min_iter: int = 0,
     precond: Optional[Callable[[array], array]] = None,
     callback: Optional[Callable[[OptimizeResult], None]] = None,
     calc_fun: bool = False,
@@ -183,6 +185,8 @@ def mmmg(  # pylint: disable=too-many-locals
         is inferior to `x0.size * tol`.
     max_iter : int, optional
         The maximum number of iterations.
+    min_iter : int, optional
+        The minimum number of iterations.
     precond : callable, optional
         A callable that must implement a preconditioner, that is `Px`. Must be a
         callable with a unique input parameter `x` and unique output like `x`.
@@ -234,7 +238,7 @@ def mmmg(  # pylint: disable=too-many-locals
         res.fun = lastgv(objv_list)
 
         # Stopping test
-        if res.grad_norm[-1] < x0.size * tol:
+        if res.grad_norm[-1] < x0.size * tol and iteration >= min_iter:
             res.success = True
             res.status = 0
             break
@@ -288,6 +292,7 @@ def mmcg(  # pylint: disable=too-many-locals
     x0: array,  # pylint: disable=invalid-name
     tol: float = 1e-4,
     max_iter: int = 500,
+    min_iter: int = 0,
     precond: Optional[Callable[[array], array]] = None,
     callback: Optional[Callable[[OptimizeResult], None]] = None,
     calc_fun: bool = False,
@@ -310,6 +315,8 @@ def mmcg(  # pylint: disable=too-many-locals
         is inferior to `x0.size * tol`.
     max_iter : int, optional
         The maximum number of iterations.
+    min_iter : int, optional
+        The minimum number of iterations.
     precond : callable, optional
         A callable that must implement a preconditioner, that is `Px`. Must be a
         callable with a unique input parameter `x` and unique output like `x`.
@@ -350,7 +357,7 @@ def mmcg(  # pylint: disable=too-many-locals
     for iteration in range(max_iter):
         # Stop test
         res.grad_norm.append(la.norm(residual))
-        if res.grad_norm[-1] < x0.size * tol:
+        if res.grad_norm[-1] < x0.size * tol and iteration >= min_iter:
             break
 
         # update
@@ -409,9 +416,9 @@ def lcg(  # pylint: disable=too-many-locals
     x0: array,  # pylint: disable=invalid-name
     tol: float = 1e-4,
     max_iter: int = 500,
+    min_iter: int = 0,
     precond: Optional[Callable[[array], array]] = None,
     callback: Optional[Callable[[OptimizeResult], None]] = None,
-    calc_fun: bool = False,
 ) -> OptimizeResult:
     """Linear Conjugate Gradient (CG) algorithm.
 
@@ -432,12 +439,11 @@ def lcg(  # pylint: disable=too-many-locals
         is inferior to `x0.size * tol`.
     max_iter : int, optional
         The maximum number of iterations.
+    min_iter : int, optional
+        The minimum number of iterations.
     callback : callable, optional
         A function that receive the `OptimizeResult` at the end of each
         iteration.
-    calc_fun: boolean, optional
-        If True, objective function is computed at each iteration with low
-        overhead. False by default. Not used by the algorithm.
 
     Returns
     -------
@@ -451,7 +457,7 @@ def lcg(  # pylint: disable=too-many-locals
 
     res.x = x0.copy().reshape((-1, 1))
 
-    second_term = np.reshape(
+    second_member = np.reshape(
         reduce(iadd, (qobjv.VtB_data for qobjv in objv_list)), (-1, 1)
     )
     constant = reduce(iadd, (c.constant for c in objv_list))
@@ -460,10 +466,10 @@ def lcg(  # pylint: disable=too-many-locals
         return reduce(iadd, (vect_call(c.hessp, arr, x0.shape) for c in objv_list))
 
     def value_residual(arr, residual):
-        return (np.sum(arr * (-second_term - residual)) + constant) / 2
+        return (np.sum(arr * (-second_member - residual)) + constant) / 2
 
     # Gradient at current x0
-    residual = second_term - hessian(res.x)
+    residual = second_member - hessian(res.x)
     direction = vect_call(precond, residual, x0.shape)
 
     res.grad_norm.append(np.sum(np.real(np.conj(residual) * direction)))
@@ -471,8 +477,7 @@ def lcg(  # pylint: disable=too-many-locals
 
     for iteration in range(max_iter):
         hessp = hessian(direction)
-        if calc_fun:
-            res.fun = value_residual(res.x, residual)
+        res.fun = value_residual(res.x, residual)
 
         # s = rᵀr / dᵀAd
         # Optimal step
@@ -483,7 +488,7 @@ def lcg(  # pylint: disable=too-many-locals
 
         # r^(i+1) = r^(i) - s * A·d
         if iteration % 50 == 0:
-            residual = second_term - hessian(res.x)
+            residual = second_member - hessian(res.x)
         else:
             residual -= step * hessp
         res.jac = -residual.reshape(x0.shape)
@@ -497,7 +502,134 @@ def lcg(  # pylint: disable=too-many-locals
         res.time.append(time.time())
 
         # Stopping condition
-        if np.sqrt(res.grad_norm[-1]) < x0.size * tol:
+        if np.sqrt(res.grad_norm[-1]) < x0.size * tol and iteration >= min_iter:
+            res.success = True
+            res.status = 0
+            break
+
+        if callback is not None:
+            callback(res)
+
+    if res.status == 0:
+        res.message = "Stopping conditions reached."
+    else:
+        res.success = False
+        res.status = 1
+        res.message = "Maximum number of iterations has been exceeded."
+    res.x = np.reshape(res.x, x0.shape)
+    res.njev = iteration + 1
+    res.nit = iteration + 1
+    res.grad_norm = list(np.sqrt(res.grad_norm))
+    res.time = list(np.asarray(res.time) - res.time[0])
+
+    return res
+
+
+def pcg(  # pylint: disable=too-many-locals
+    normalp: Callable[[array], array],
+    secondm: array,
+    x0: array,  # pylint: disable=invalid-name
+    tol: float = 1e-4,
+    max_iter: int = 500,
+    min_iter: int = 0,
+    precond: Optional[Callable[[array], array]] = None,
+    callback: Optional[Callable[[OptimizeResult], None]] = None,
+) -> OptimizeResult:
+    """Preconditionnate Conjugate Gradient (CG) algorithm.
+
+    Preconditionnate Conjugate Gradient for solving SDP linear system
+
+    .. math::
+        A x = b
+
+    where :math:`A` is a SDP matrix and :math:`b` a vector. It comes from the
+    gradient cancelling of quadratic criterion
+
+    .. math::
+        J(x) = \frac{1}{2} x^T A x - 2 x^T b + c.
+
+    Parameters
+    ----------
+    normalp : callable
+        A callable that apply the matrix `A` on a vector `x`. Must accept an
+        array and return an array, both of any shape.
+    secondm : array
+        The second member `b`. Can be of any shape.
+    x0 : ndarray
+        The initial point.
+    precond : callable, optional
+        A callable that must implement a preconditioner, that is `Px`. Must be a
+        callable with a unique input parameter `x` and unique output like `x`.
+    tol : float, optional
+        The stopping tolerance. The algorithm is stopped when the gradient norm
+        is inferior to `x0.size * tol`.
+    max_iter : int, optional
+        The maximum number of iterations.
+    min_iter : int, optional
+        The minimum number of iterations.
+    callback : callable, optional
+        A function that receive the `OptimizeResult` at the end of each
+        iteration.
+
+    Returns
+    -------
+    result : OptimizeResult
+
+    Notes
+    -----
+    This is the same algorithm than `lcg` with a different API.
+
+    """
+
+    if precond is None:
+        precond = lambda x: x
+    res = OptimizeResult()
+
+    res.x = x0.copy().reshape((-1, 1))
+
+    second_member = np.reshape(secondm, (-1, 1))
+
+    def hessian(arr):
+        return vect_call(normalp, arr, x0.shape)
+
+    def value_residual(arr, residual):
+        return np.sum(arr * (-second_member - residual)) / 2
+
+    # Gradient at current x0
+    residual = second_member - hessian(res.x)
+    direction = vect_call(precond, residual, x0.shape)
+
+    res.grad_norm.append(np.sum(np.real(np.conj(residual) * direction)))
+    res.time.append(time.time())
+
+    for iteration in range(max_iter):
+        hessp = hessian(direction)
+        res.fun = value_residual(res.x, residual)
+
+        # s = rᵀr / dᵀAd
+        # Optimal step
+        step = res.grad_norm[-1] / np.sum(np.real(np.conj(direction) * hessp))
+
+        # Descent x^(i+1) = x^(i) + s*d
+        res.x += step * direction
+
+        # r^(i+1) = r^(i) - s * A·d
+        if iteration % 50 == 0:
+            residual = second_member - hessian(res.x)
+        else:
+            residual -= step * hessp
+        res.jac = -residual.reshape(x0.shape)
+
+        # Conjugate direction with preconditionner
+        secant = vect_call(precond, residual, x0.shape)
+        res.grad_norm.append(np.sum(np.real(np.conj(residual) * secant)))
+        direction = secant + (res.grad_norm[-1] / res.grad_norm[-2]) * direction
+
+        res.diff.append(np.sum(step * direction) ** 2)
+        res.time.append(time.time())
+
+        # Stopping condition
+        if np.sqrt(res.grad_norm[-1]) < x0.size * tol and iteration >= min_iter:
             res.success = True
             res.status = 0
             break
@@ -531,7 +663,15 @@ def vect_call(func: Callable[[array], array], point: array, shape: Tuple) -> arr
 
 # Not used in the module, only provided for user convenience
 def vectorize(shape: Tuple) -> Callable:
-    """Return a decorator to vectorize input and output given `shape`"""
+    """Return a decorator to vectorize input and output given `shape`.
+
+    Suppose you have a function (a gradient typically) that demands an array
+    with a specific shape, say (N, M) and return an array with another shape,
+    say (P, Q). But you use an algorithm (optimize from scipy for instance) that
+    demands vector only of shape (R, ) and (T,). This decorator factory return a
+    decorator that do the job conversion
+
+    """
 
     def decorator(func: Callable[[array], array]) -> Callable[[array], array]:
         """A decorator to vectorize input and output"""
@@ -1252,7 +1392,7 @@ class Square(Loss):
         super().__init__(inf=1, convex=True, coercive=True)
 
     def value(self, point: array) -> array:
-        return point**2 / 2
+        return point ** 2 / 2
 
     def gradient(self, point: array) -> array:
         return point
@@ -1283,7 +1423,7 @@ class Huber(Loss):
     def value(self, point: array) -> array:
         return np.where(
             np.abs(point) <= self.delta,
-            point**2 / 2,
+            point ** 2 / 2,
             self.delta * (np.abs(point) - self.delta / 2),
         )
 
@@ -1319,10 +1459,10 @@ class Hyperbolic(Loss):
         self.delta = delta
 
     def value(self, point: array) -> array:
-        return self.delta**2 * (np.sqrt(1 + (point**2) / (self.delta**2)) - 1)
+        return self.delta ** 2 * (np.sqrt(1 + (point ** 2) / (self.delta ** 2)) - 1)
 
     def gradient(self, point: array) -> array:
-        return point / np.sqrt(1 + (point**2) / self.delta**2)
+        return point / np.sqrt(1 + (point ** 2) / self.delta ** 2)
 
     def __repr__(self):
         return f"""{type(self)}
@@ -1347,14 +1487,14 @@ class HebertLeahy(Loss):
 
     def __init__(self, delta: float):
         """The Hebert & Leahy loss."""
-        super().__init__(inf=2 / delta**2, convex=False, coercive=True)
+        super().__init__(inf=2 / delta ** 2, convex=False, coercive=True)
         self.delta = delta
 
     def value(self, point: array) -> array:
-        return np.log(1 + point**2 / self.delta**2)
+        return np.log(1 + point ** 2 / self.delta ** 2)
 
     def gradient(self, point: array) -> array:
-        return 2 * point / (self.delta**2 + point**2)
+        return 2 * point / (self.delta ** 2 + point ** 2)
 
     def __repr__(self):
         return f"""{type(self)}
@@ -1378,14 +1518,14 @@ class GemanMcClure(Loss):
 
     def __init__(self, delta: float):
         r"""The Geman & Mc Clure loss."""
-        super().__init__(1 / (delta**2), convex=False, coercive=False)
+        super().__init__(1 / (delta ** 2), convex=False, coercive=False)
         self.delta = delta
 
     def value(self, point: array) -> array:
-        return point**2 / (2 * self.delta**2 + point**2)
+        return point ** 2 / (2 * self.delta ** 2 + point ** 2)
 
     def gradient(self, point: array) -> array:
-        return 4 * point * self.delta**2 / (2 * self.delta**2 + point**2) ** 2
+        return 4 * point * self.delta ** 2 / (2 * self.delta ** 2 + point ** 2) ** 2
 
     def __repr__(self):
         return f"""{type(self)}
@@ -1409,14 +1549,14 @@ class TruncSquareApprox(Loss):
 
     def __init__(self, delta: array):
         """The truncated square approximation."""
-        super().__init__(inf=1 / (delta**2), convex=False, coercive=False)
+        super().__init__(inf=1 / (delta ** 2), convex=False, coercive=False)
         self.delta = delta
 
     def value(self, point: array) -> array:
-        return 1 - np.exp(-(point**2) / (2 * self.delta**2))
+        return 1 - np.exp(-(point ** 2) / (2 * self.delta ** 2))
 
     def gradient(self, point: array) -> array:
-        return point / (self.delta**2) * np.exp(-(point**2) / (2 * self.delta**2))
+        return point / (self.delta ** 2) * np.exp(-(point ** 2) / (2 * self.delta ** 2))
 
     def __repr__(self):
         return f"""{type(self)}
